import { useRef, useState, useEffect, Fragment, createElement } from "react";
import * as ReactDom from "react-dom";
const slotToReactComponent = (slotRenderFn) => (props) => {
    const ref = useRef(null);
    const hfcSlot = useRef(null);
    useEffect(() => {
        if (hfcSlot.current !== null) {
            hfcSlot.current.args = props.args;
            hfcSlot.current.changed?.();
        }
    });
    useEffect(() => {
        hfcSlot.current = {
            args: props.args,
            target: ref.current,
        };
        slotRenderFn(hfcSlot.current);
        return () => {
            // unmount slot
            hfcSlot.current.removed?.();
        };
    }, []);
    return createElement(props.tag || "div", { ...props, ref, tag: undefined, args: undefined }, null);
};
function toReactProps(props) {
    const reactProps = { ...props.attrs, ...props.events, ...props._ };
    if (props.slots) {
        for (const key in props.slots) {
            const slotRenderFn = props.slots[key];
            if (!slotRenderFn)
                continue;
            reactProps[key] = slotToReactComponent(slotRenderFn);
            reactProps[key]._render = slotRenderFn;
        }
    }
    return reactProps;
}
const portals = new Map();
let uuid = 0;
let forceUpdateRoot = () => { };
export const rootElement = document.createElement("div");
function HFCReactRoot() {
    const [, update] = useState(0);
    forceUpdateRoot = () => update((n) => n + 1);
    return createElement(Fragment, null, Array.from(portals).map(([key, { container, node }]) => ReactDom.createPortal(node, container, key)));
}
const createRoot = ReactDom.createRoot;
if (createRoot) {
    // hijack createRoot error log
    const log = console.error;
    console.error = () => { };
    const root = createRoot(rootElement);
    console.error = log;
    root.render(createElement(HFCReactRoot));
}
else {
    ReactDom.render(createElement(HFCReactRoot), rootElement);
}
export function toHFC(Comp, opts) {
    const HFC = (initProps) => {
        const key = "k" + uuid++;
        const currentProps = {
            attrs: initProps.attrs || {},
            events: initProps.events || {},
            slots: initProps.slots || {},
            _: initProps._ || {},
        };
        let forceUpdateWrapper;
        function Wrapper() {
            const [, update] = useState(0);
            forceUpdateWrapper = () => update((n) => n + 1);
            const props = toReactProps(currentProps);
            return createElement(Comp, props);
        }
        return {
            connected(container) {
                if (opts.connected)
                    opts.connected(container);
                portals.set(key, {
                    container,
                    node: createElement(Wrapper),
                });
                forceUpdateRoot();
            },
            changed(props, partial) {
                if (opts.changed)
                    opts.changed(props, partial);
                if (props.attrs) {
                    currentProps.attrs = partial
                        ? { ...currentProps.attrs, ...props.attrs }
                        : props.attrs;
                }
                if (props.events) {
                    currentProps.events = partial
                        ? { ...currentProps.events, ...props.events }
                        : props.events;
                }
                if (props.slots) {
                    currentProps.slots = partial
                        ? { ...currentProps.slots, ...props.slots }
                        : props.slots;
                }
                if (props._) {
                    currentProps._ = partial
                        ? { ...currentProps._, ...props._ }
                        : props._;
                }
                forceUpdateWrapper();
            },
            disconnected() {
                if (opts.disconnected)
                    opts.disconnected();
                portals.delete(key);
                forceUpdateRoot();
            },
        };
    };
    HFC.tag = opts.tag;
    HFC.hfc = opts.hfc;
    HFC.ver = opts.ver;
    HFC.names = opts.names;
    return HFC;
}
export function toHFCReact(Comp, opts) {
    const slotKeys = new Set(opts.names[2]);
    return function HFCReact(props) {
        const container = useRef(null);
        const _props = {};
        for (const key in props) {
            const value = props[key];
            // slot key
            if (slotKeys.has(key)) {
                _props[key] = toHfcSlot(value);
            }
            else {
                _props[key] = value;
            }
        }
        useEffect(() => {
            if (opts.connected)
                opts.connected(container.current);
            return () => {
                if (opts.disconnected)
                    opts.disconnected();
            };
        }, []);
        return createElement(opts.tag, { ref: container }, createElement(Comp, _props));
    };
}
function toHfcSlot(Comp) {
    return (props) => {
        return createElement(props.tag || "div", props, createElement(Comp, props.args));
    };
}
